/*! ********************************************
 * EffectorExt.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_EffectorExt_h
#include "EffectorExt.hpp"
#endif

#include <utility>

namespace sora {
namespace graybox {
namespace efx {
void EFXStereoDelay::setSampleRate(float rate)
{ 
	_sampleRate = rate;
	_delayTimeL = _delayL / _sampleRate;
	_delayTimeR = _delayR / _sampleRate;
}
float EFXStereoDelay::setDelayTimeL(float val)
{ 
	int _frame = int(val * _sampleRate);
	if (_frame > _ringBufferMask)_frame = _ringBufferMask;
	_delayL = _frame;
	_rpL = _wp + (_ringBufferMask - _delayL) & _ringBufferMask;
	return _delayTimeL = _frame / _sampleRate;
}
float EFXStereoDelay::setDelayTimeR(float val)
{
	int _frame = int(val * _sampleRate);
	if (_frame > _ringBufferMask)_frame = _ringBufferMask;
	_delayR = _frame;
	_rpR = _wp + (_ringBufferMask - _delayR) & _ringBufferMask;
	return _delayTimeR = _frame / _sampleRate;
}
void EFXStereoDelay::setDelayFrameL(int32 val)
{ 
	_delayL = val;
	if (_delayL > _ringBufferMask)_delayL = _ringBufferMask;
	_rpL = _wp + (_ringBufferMask - _delayL) & _ringBufferMask;
	_delayTimeL = _delayL / _sampleRate;
}
void EFXStereoDelay::setDelayFrameR(int32 val)
{
	_delayR = val;
	if (_delayR > _ringBufferMask)_delayR = _ringBufferMask;
	_rpR = _wp + (_ringBufferMask - _delayR) & _ringBufferMask;
	_delayTimeR = _delayR / _sampleRate;
}
void EFXStereoDelay::setEffectParam(float mix,
	float delayTimeL, float feedBackLevelL,
	float delayTimeR, float feedBackLevelR)
{
	_mix = mix;
	_antiMix = 1.0f - _mix;
	setDelayTimeL(delayTimeL);
	setDelayTimeR(delayTimeR);
	_feedBackL = feedBackLevelL;
	_feedBackR = feedBackLevelR;
}
// Base - EffectorBase ---
void EFXStereoDelay::applyEffect(float* const& outL, float* const& outR, const int32& startFrame, const int32& sampleFrames)
{
	int _endFrame = startFrame + sampleFrames;
	float _delay;
	for (int i = startFrame; i < _endFrame; ++i,
		_rpL = _rpL + 1 & _ringBufferMask,
		_rpR = _rpR + 1 & _ringBufferMask,
		_wp  = _wp  + 1 & _ringBufferMask)
	{
	// Channel L --->
		_delay = _delayBufferL[_rpL];
		_delayBufferL[_wp] = outL[i] + _delay * _feedBackL;
		outL[i] = outL[i] * _antiMix + _delay * _mix;	// outout
	// Channel R --->
		_delay = _delayBufferR[_rpR];
		_delayBufferR[_wp] = outR[i] + _delay * _feedBackR;
		outR[i] = outR[i] * _antiMix + _delay * _mix;	// outout
	}
}
void EFXStereoDelay::resetParameters()
{
	clearBuffer();
}
void EFXStereoDelay::clearBuffer()
{
	for (int i = 0; i < _ringBufferSize; ++i)_delayBufferL[i] = _delayBufferR[i] = 0.0f;
}
void EFXStereoDelay::initialize()
{
	clearBuffer();
	setEffectParam();
}
// Interface - IfGrayBoxModule -----
void EFXStereoDelay::updateSampleRate(float sampleRate){ setSampleRate(sampleRate); }

void EFXStereoDelay::resetModule(){ initialize(); }
void EFXStereoDelay::initiModule(){ initialize(); }
void EFXStereoDelay::connectModule(GrayBox* host)
{ 
	clearBuffer();
	setSampleRate(host->getSampleRate());
}
void EFXStereoDelay::disconnectModule(){ initialize(); }
void EFXStereoDelay::updateBypassStatus(bool){ initialize(); }

void EFXStereoDelay::startProcess(){ initialize(); }
void EFXStereoDelay::stopProcess(){ initialize(); }
void EFXStereoDelay::suspendProcess(){ initialize(); }
void EFXStereoDelay::resumeProcess(){ initialize(); }
void EFXStereoDelay::process(ProcessData* procData)
{
	// inline --- applyEffect()
	int i = procData->offsetFrames,
	    _endFrame = i + procData->sampleFrames;
	float _delay;
	float*& _outL = procData->audioOut[0];
	float*& _outR = procData->audioOut[1];
	for ( ; i < _endFrame; ++i,
		_rpL = _rpL + 1 & _ringBufferMask,
		_rpR = _rpR + 1 & _ringBufferMask,
		_wp  = _wp  + 1 & _ringBufferMask)
	{
	// Channel L ---------
		_delay = _delayBufferL[_rpL];
		_delayBufferL[_wp] = _outL[i] + _delay * _feedBackL;
		_outL[i] = _outL[i] * _antiMix + _delay * _mix;	// outout
	// Channel R ---------
		_delay = _delayBufferR[_rpR];
		_delayBufferR[_wp] = _outR[i] + _delay * _feedBackR;
		_outR[i] = _outR[i] * _antiMix + _delay * _mix;	// outout
	}
}
// Constructor
EFXStereoDelay::EFXStereoDelay() : EffectorBase::EffectorBase()
{
	// bufferSize
	_delayBufferL = new float[_ringBufferSize];
	_delayBufferR = new float[_ringBufferSize];
	// reset
	clearBuffer();
	setEffectParam();
}
EFXStereoDelay::EFXStereoDelay(int32 bufferSizeBits) : EffectorBase::EffectorBase()
{
	// bufferSize
	if (bufferSizeBits > 20) bufferSizeBits = 20;
	else if (bufferSizeBits < 4) bufferSizeBits = 4;
	_ringBufferSizeBits = bufferSizeBits;
	_ringBufferSize = 1 << _ringBufferSizeBits;
	_ringBufferMask = _ringBufferSize - 1;
	_delayBufferL = new float[_ringBufferSize];
	_delayBufferR = new float[_ringBufferSize];
	// reset
	clearBuffer();
	setEffectParam();
}
EFXStereoDelay::EFXStereoDelay(const EFXStereoDelay & x)
{
	// base
	_mix = x._mix;
	_antiMix = x._antiMix;
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	_delayBufferL = new float[_ringBufferSize];
	_delayBufferR = new float[_ringBufferSize];
	for (int i = 0; i < _ringBufferSize; ++i)
	{
		_delayBufferL[i] = x._delayBufferL[i];
		_delayBufferR[i] = x._delayBufferR[i];
	}
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBackL = x._feedBackL;
	_feedBackR = x._feedBackR;
	_rpL = x._rpL;
	_rpR = x._rpR;
	_delayL = x._delayL;
	_delayR = x._delayR;
	_delayTimeL = x._delayTimeL;
	_delayTimeR = x._delayTimeR;
}
EFXStereoDelay & EFXStereoDelay::operator=(const EFXStereoDelay & x)
{
	if (this == &x)return *this;
	// base
	_mix = x._mix;
	_antiMix = x._antiMix;
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	_delayBufferL = new float[_ringBufferSize];
	_delayBufferR = new float[_ringBufferSize];
	for (int i = 0; i < _ringBufferSize; ++i)
	{
		_delayBufferL[i] = x._delayBufferL[i];
		_delayBufferR[i] = x._delayBufferR[i];
	}
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBackL = x._feedBackL;
	_feedBackR = x._feedBackR;
	_rpL = x._rpL;
	_rpR = x._rpR;
	_delayL = x._delayL;
	_delayR = x._delayR;
	_delayTimeL = x._delayTimeL;
	_delayTimeR = x._delayTimeR;
	return *this;
}
EFXStereoDelay::EFXStereoDelay(EFXStereoDelay && x) noexcept : EffectorBase(std::move(x))
{
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	_delayBufferL = x._delayBufferL; x._delayBufferL = nullptr;
	_delayBufferR = x._delayBufferR; x._delayBufferR = nullptr;
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBackL = x._feedBackL;
	_feedBackR = x._feedBackR;
	_rpL = x._rpL;
	_rpR = x._rpR;
	_delayL = x._delayL;
	_delayR = x._delayR;
	_delayTimeL = x._delayTimeL;
	_delayTimeR = x._delayTimeR;
	
}
EFXStereoDelay & EFXStereoDelay::operator=(EFXStereoDelay && x) noexcept
{
	if (this == &x)return *this;
	// base
	_mix = x._mix;
	_antiMix = x._antiMix;
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	if (_delayBufferL != nullptr)delete[] _delayBufferL;
	_delayBufferL = x._delayBufferL; x._delayBufferL = nullptr;
	if (_delayBufferR != nullptr)delete[] _delayBufferR;
	_delayBufferR = x._delayBufferR; x._delayBufferR = nullptr;
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBackL = x._feedBackL;
	_feedBackR = x._feedBackR;
	_rpL = x._rpL;
	_rpR = x._rpR;
	_delayL = x._delayL;
	_delayR = x._delayR;
	_delayTimeL = x._delayTimeL;
	_delayTimeR = x._delayTimeR;
	
	return *this;
}
EFXStereoDelay::~EFXStereoDelay()
{
	if (_delayBufferL != nullptr) delete[] _delayBufferL;
	if (_delayBufferR != nullptr) delete[] _delayBufferR;
	_delayBufferL = nullptr;
	_delayBufferR = nullptr;
}
void EFXStereoRevervb::setSampleRate(float rate)
{ 
	_sampleRate = rate;
	_delayTime1 = _delay1 / _sampleRate;
	_delayTime2 = _delay2 / _sampleRate;
}
float EFXStereoRevervb::setDelayTime1(float val)
{ 
	int _frame = int(val * _sampleRate);
	if (_frame > _ringBufferMask)_frame = _ringBufferMask;
	_delay1 = _frame;
	_rp1 = _wp + (_ringBufferMask - _delay1) & _ringBufferMask;
	return _delayTime1 = _frame / _sampleRate;
}
float EFXStereoRevervb::setDelayTime2(float val)
{
	int _frame = int(val * _sampleRate);
	if (_frame > _ringBufferMask)_frame = _ringBufferMask;
	_delay2 = _frame;
	_rp2 = _wp + (_ringBufferMask - _delay2) & _ringBufferMask;
	return _delayTime2 = _frame / _sampleRate;
}
void EFXStereoRevervb::setDelayFrame1(int32 val)
{ 
	_delay1 = val;
	if (_delay1 > _ringBufferMask)_delay1 = _ringBufferMask;
	_rp1 = _wp + (_ringBufferMask - _delay1) & _ringBufferMask;
	_delayTime1 = _delay1 / _sampleRate;
}
void EFXStereoRevervb::setDelayFrame2(int32 val)
{
	_delay2 = val;
	if (_delay2 > _ringBufferMask)_delay2 = _ringBufferMask;
	_rp2 = _wp + (_ringBufferMask - _delay2) & _ringBufferMask;
	_delayTime2 = _delay2 / _sampleRate;
}
void EFXStereoRevervb::setEffectParam(float mix,
	float delayTime1, float feedBackLevel1,
	float delayTime2, float feedBackLevel2)
{
	_mix = mix;
	_antiMix = 1.0f - _mix;
	setDelayTime1(delayTime1);
	setDelayTime2(delayTime2);
	_feedBack1 = feedBackLevel1;
	_feedBack2 = feedBackLevel2;
}
// Base - EffectorBase ---
void EFXStereoRevervb::applyEffect(float* const& outL, float* const& outR, const int32& startFrame, const int32& sampleFrames)
{
	int _endFrame = startFrame + sampleFrames;
	float _v;
	for (int i = startFrame; i < _endFrame; ++i,
		_rp1 = _rp1 + 1 & _ringBufferMask,
		_rp2 = _rp2 + 1 & _ringBufferMask,
		_wp  = _wp  + 1 & _ringBufferMask)
	{
	// Channel L --->
		_v  = _delayBuffer1[_rp1] * _feedBack1;
		_v += _delayBuffer1[_rp2] * _feedBack2;
		_delayBuffer1[_wp] = outL[i] + _v;
		outL[i] = outL[i] * _antiMix + _v * _mix; // outout
	// Channel R --->
		_v  = _delayBuffer2[_rp1] * _feedBack1;
		_v += _delayBuffer2[_rp2] * _feedBack2;
		_delayBuffer2[_wp] = outR[i] + _v;
		outR[i] = outR[i] * _antiMix + _v * _mix; // outout
	}
}
void EFXStereoRevervb::resetParameters()
{
	clearBuffer();
}
void EFXStereoRevervb::clearBuffer()
{
	for (int i = 0; i < _ringBufferSize; ++i)_delayBuffer1[i] = _delayBuffer2[i] = 0.0f;
}
void EFXStereoRevervb::initialize()
{
	clearBuffer();
	setEffectParam();
}
// Interface - IfGrayBoxModule -----
void EFXStereoRevervb::updateSampleRate(float sampleRate){ setSampleRate(sampleRate); }

void EFXStereoRevervb::resetModule(){ initialize(); }
void EFXStereoRevervb::initiModule(){ initialize(); }
void EFXStereoRevervb::connectModule(GrayBox* host)
{ 
	clearBuffer();
	setSampleRate(host->getSampleRate());
}
void EFXStereoRevervb::disconnectModule(){ initialize(); }
void EFXStereoRevervb::updateBypassStatus(bool){ initialize(); }

void EFXStereoRevervb::startProcess(){ initialize(); }
void EFXStereoRevervb::stopProcess(){ initialize(); }
void EFXStereoRevervb::suspendProcess(){ initialize(); }
void EFXStereoRevervb::resumeProcess(){ initialize(); }
void EFXStereoRevervb::process(ProcessData* procData)
{
	// inline --- applyEffect()
	int i = procData->offsetFrames,
	    _endFrame = i + procData->sampleFrames;
	float _v;
	float*& _outL = procData->audioOut[0];
	float*& _outR = procData->audioOut[1];
	for ( ; i < _endFrame; ++i,
		_rp1 = _rp1 + 1 & _ringBufferMask,
		_rp2 = _rp2 + 1 & _ringBufferMask,
		_wp  = _wp  + 1 & _ringBufferMask)
	{
	// Channel L --->
		_v  = _delayBuffer1[_rp1] * _feedBack1;
		_v += _delayBuffer1[_rp2] * _feedBack2;
		_delayBuffer1[_wp] = _outL[i] + _v;
		_outL[i] = _outL[i] * _antiMix + _v * _mix; // outout
	// Channel R --->
		_v  = _delayBuffer2[_rp1] * _feedBack1;
		_v += _delayBuffer2[_rp2] * _feedBack2;
		_delayBuffer2[_wp] = _outR[i] + _v;
		_outR[i] = _outR[i] * _antiMix + _v * _mix; // outout
	}
}
// Constructor
EFXStereoRevervb::EFXStereoRevervb() : EffectorBase::EffectorBase()
{
	// bufferSize
	_delayBuffer1 = new float[_ringBufferSize];
	_delayBuffer2 = new float[_ringBufferSize];
	// reset
	clearBuffer();
	setEffectParam();
}
// Constructor
EFXStereoRevervb::EFXStereoRevervb(int32 bufferSizeBits) : EffectorBase::EffectorBase()
{
	// bufferSize
	if (bufferSizeBits > 20) bufferSizeBits = 20;
	else if (bufferSizeBits < 4) bufferSizeBits = 4;
	_ringBufferSizeBits = bufferSizeBits;
	_ringBufferSize = 1 << _ringBufferSizeBits;
	_ringBufferMask = _ringBufferSize - 1;
	_delayBuffer1 = new float[_ringBufferSize];
	_delayBuffer2 = new float[_ringBufferSize];
	// reset
	clearBuffer();
	setEffectParam();
}
EFXStereoRevervb::EFXStereoRevervb(const EFXStereoRevervb & x)
{
	// base
	_mix = x._mix;
	_antiMix = x._antiMix;
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	_delayBuffer1 = new float[_ringBufferSize];
	_delayBuffer2 = new float[_ringBufferSize];
	for (int i = 0; i < _ringBufferSize; ++i)
	{
		_delayBuffer1[i] = x._delayBuffer1[i];
		_delayBuffer2[i] = x._delayBuffer2[i];
	}
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBack1 = x._feedBack1;
	_feedBack2 = x._feedBack2;
	_rp1 = x._rp1;
	_rp2 = x._rp2;
	_delay1 = x._delay1;
	_delay2 = x._delay2;
	_delayTime1 = x._delayTime1;
	_delayTime2 = x._delayTime2;
}
EFXStereoRevervb & EFXStereoRevervb::operator=(const EFXStereoRevervb & x)
{
	if (this == &x)return *this;
	// base
	_mix = x._mix;
	_antiMix = x._antiMix;
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	_delayBuffer1 = new float[_ringBufferSize];
	_delayBuffer2 = new float[_ringBufferSize];
	for (int i = 0; i < _ringBufferSize; ++i)
	{
		_delayBuffer1[i] = x._delayBuffer1[i];
		_delayBuffer2[i] = x._delayBuffer2[i];
	}
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBack1 = x._feedBack1;
	_feedBack2 = x._feedBack2;
	_rp1 = x._rp1;
	_rp2 = x._rp2;
	_delay1 = x._delay1;
	_delay2 = x._delay2;
	_delayTime1 = x._delayTime1;
	_delayTime2 = x._delayTime2;
	return *this;
}
EFXStereoRevervb::EFXStereoRevervb(EFXStereoRevervb && x) noexcept : EffectorBase(std::move(x))
{
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	_delayBuffer1 = x._delayBuffer1; x._delayBuffer1 = nullptr;
	_delayBuffer2 = x._delayBuffer2; x._delayBuffer2 = nullptr;
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBack1 = x._feedBack1;
	_feedBack2 = x._feedBack2;
	_rp1 = x._rp1;
	_rp2 = x._rp2;
	_delay1 = x._delay1;
	_delay2 = x._delay2;
	_delayTime1 = x._delayTime1;
	_delayTime2 = x._delayTime2;
	
}
EFXStereoRevervb & EFXStereoRevervb::operator=(EFXStereoRevervb && x) noexcept
{
	if (this == &x)return *this;
	// base
	_mix = x._mix;
	_antiMix = x._antiMix;
	// Ring Buffer
	_ringBufferSizeBits = x._ringBufferSizeBits;
	_ringBufferSize = x._ringBufferSize;
	_ringBufferMask = x._ringBufferMask;
	if (_delayBuffer1 != nullptr)delete[] _delayBuffer1;
	_delayBuffer1 = x._delayBuffer1; x._delayBuffer1 = nullptr;
	if (_delayBuffer2 != nullptr)delete[] _delayBuffer2;
	_delayBuffer2 = x._delayBuffer2; x._delayBuffer2 = nullptr;
	// 
	_sampleRate = x._sampleRate;
	// Delay
	_wp = x._wp;
	_feedBack1 = x._feedBack1;
	_feedBack2 = x._feedBack2;
	_rp1 = x._rp1;
	_rp2 = x._rp2;
	_delay1 = x._delay1;
	_delay2 = x._delay2;
	_delayTime1 = x._delayTime1;
	_delayTime2 = x._delayTime2;
	
	return *this;
}
EFXStereoRevervb::~EFXStereoRevervb()
{
	if (_delayBuffer1 != nullptr) delete[] _delayBuffer1;
	if (_delayBuffer2 != nullptr) delete[] _delayBuffer2;
	_delayBuffer1 = nullptr;
	_delayBuffer2 = nullptr;
}
}}}