/*! ********************************************
 * SampleData.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SampleData_h
#define INCLUDE_SampleData_h

#ifndef INCLUDE_SynthData_h
#include "SynthData.hpp"
#endif

namespace sora {
namespace graybox {
namespace synth {
class SamplerPreset
{
public:
	//int32* parameterPointer[16];
	char toneName[17] = { "                " };	// ASCII 16Chars 
	int32 smplNum = 0;
	int32 pan = 64;
	int32 octv = 8;     // 0 - 16     // center 8, octave = 1
	int32 tune = 8192;  // 0 - 16383  // center 8192, octave = 8192
	int32 tl = 0;       // 0 - 127
	int32 ar = 127;     // 0 - 127
	int32 dr = 0;       // 0 - 127
	int32 sl = 127;     // 0 - 127
	int32 rr = 120;     // 0 - 127

	void copy(const SamplerPreset& preset);
	void reset();
};
class SampleData
{
	/*
		|------------- Wave Rage ------------|

		|   |          |      |        |     |
		|   |  /\    /\|    /\|  /\    |     |
		|---|\/  \  /  |\  /  |\/  \/--|-----|
		|   |     \/   | \/   |        |     |
		|   |          |      |        |     |

		|   |          |      |        |     |
		|   |          |      |        |PlayBack End (SoundOff)
		|   |          |      |Release (KeyOff)
		|   |          |      |Loop End
		|   |          |Loop Start
		|   |PlayBack Start (KeyOn)


		- sampleFrames -
		samples : [L,R][L,R]...
		frames    [ 0 ][ 1 ]...
	*/
private:
	static constexpr double _RES = float(-SynthData::LOG2VT_DYN_RES);
	static constexpr double _BOADER = 1.0 / (1 << SynthData::LOG2VT_DYN_BITS);
	static constexpr int32 _BOTTOM = SynthData::LOG2VT_SIZE;
	static constexpr float _POW2T_OCTV_RESf = float(SynthData::POW2T_OCTV_RES);
	static int32 _log2VTindex(double val);

	bool _created = false;
	float _sampleRate = 0.0f;
	float _recPitch = 0.0f;
public:
	int32* sampleData = nullptr;  // Indexed Sample Data
	int32 sampleFrames = 0;
	int32 channelCount = 0;
	// Synth Param
	int32 pgSample = 0;  // SampleRate
	int32 pgPitch  = 0;  // RecPitch
	int32 pgKey    = 0;  // SampleRate - RecPitch
	int32 plbkS    = 0;  // PlayBack Start Frame [plbkS >= 0]
	int32 plbkE    = 0;  // PlayBack End Frame   [plbkE <= sampleFrames]
	int32 loopE    = 0;  // Loop End Frame       [loopE <= plbkE]
	int32 loopF    = 0;  // Loop Frames          [val = loopE - lpStart]
	int32 loopFk   = 0;  // Loop Frames          * after -> keyOff
// getter, setter
	int32 getSampleFrames() { return sampleFrames; }
	int32 getChannelCount() { return channelCount; }
	float getSampleRate() { return _sampleRate; }
	void setSampleRate(float sampleRate);
	float getRecPitch(){ return _recPitch; }
	void setRecPitch(float pitch);
	int32 getPlayBackStartPoint() { return plbkS; }
	int32 getPlayBackEndPoint() { return plbkE; }
	int32 getLoopStartPoint() { return (loopF > 0)? loopE - loopF : 0; }
	int32 getLoopEndPoint() { return (loopF > 0)? loopE : 0; }
	bool getLoopable() { return loopF > 0; }
// Method
	void clear();
	void copy(const SampleData& from, bool deepCopy = false);
	void setPlaybackRange(int32 start, int32 end);
	void resetLoopPoints(int32 start = 0, int32 end = 0);
	void loadPCMSamples(int8* wave, int32 sampleFrames, int32 ch);
	void loadPCMSamples(uint8* wave, int32 sampleFrames, int32 ch);
	void loadPCMSamples(int16* wave, int32 sampleFrames, int32 ch);
	void loadPCMSamples(uint16* wave, int32 sampleFrames, int32 ch);
	void loadPCMSamples(int32* wave, int32 sampleFrames, int32 ch);
	void loadPCMSamples(uint32* wave, int32 sampleFrames, int32 ch);
	void loadPCMSamples(float* wave, int32 sampleFrames, int32 ch);
	void loadPCMData(uint8* dat, int32 sampleFrames, int32 ch, int32 bitRate, bool sign, bool bigEndian);
// Constructor
public:
	SampleData() = default;
// Copy Constructor
	SampleData(const SampleData&);
	SampleData& operator = (const SampleData&);
// Move Constructor
	SampleData(SampleData&&)noexcept;
	SampleData& operator = (SampleData&&)noexcept;
// Destructor
	~SampleData();
};
}}}
#endif